@@ -1,17 +1,19 @@
 import logging
-import pathlib
 import threading
 import datetime
 import time
+from collections import defaultdict
 from functools import wraps
-from typing import Dict, Iterable, Optional, Union, List
+from typing import Dict, Iterable, Optional, Union, List, Tuple
+import math
+import re
 
 from libbs.api.utils import progress_bar
 from libbs.artifacts import (
     Artifact,
     Function, FunctionHeader, StackVariable,
     Comment, GlobalVariable, Patch,
-    Enum, Struct, FunctionArgument, StructMember
+    Enum, Struct, FunctionArgument, StructMember, Typedef, Segment
 )
 from libbs.api import DecompilerInterface
 from libbs.api.type_parser import CType
@@ -21,8 +23,9 @@ from binsync.core.state import State
 from binsync.core.user import User
 from binsync.configuration import BinSyncBSConfig
 
-_l = logging.getLogger(name=__name__)
+from wordfreq import word_frequency
 
+_l = logging.getLogger(name=__name__)
 
 #
 # State Checking Decorators
@@ -38,13 +41,6 @@ def init_checker(f):
     return _init_check
 
 
-def fill_event(f):
-    @wraps(f)
-    def _fill_event(self: "BSController", *args, **kwargs):
-        return self.fill_event_handler(f, *args, **kwargs)
-
-    return _fill_event
-
 #
 # Description Constants
 #
@@ -84,7 +80,7 @@ class BSController:
 
     """
     CHANGE_WATCHERS = (
-        FunctionHeader, StackVariable, Comment, GlobalVariable, Enum, Struct
+        FunctionHeader, StackVariable, Comment, GlobalVariable, Enum, Struct, Typedef, Segment
     )
 
     ARTIFACT_SET_MAP = {
@@ -94,7 +90,9 @@ class BSController:
         Comment: State.set_comment,
         GlobalVariable: State.set_global_var,
         Struct: State.set_struct,
-        Enum: State.set_enum
+        Enum: State.set_enum,
+        Typedef: State.set_typedef,
+        Segment: State.set_segment
     }
 
     ARTIFACT_GET_MAP = {
@@ -112,12 +110,18 @@ class BSController:
         (Struct, GET_MANY): State.get_structs,
         Enum: State.get_enum,
         (Enum, GET_MANY): State.get_enums,
+        Typedef: State.get_typedef,
+        (Typedef, GET_MANY): State.get_typedefs,
+        Segment: State.get_segment,
+        (Segment, GET_MANY): State.get_segments
     }
 
     DEFAULT_SEMAPHORE_SIZE = 100
 
-    def __init__(self, decompiler_interface: DecompilerInterface = None, headless=False, auto_commit=True,
-                 reload_time=10, **kwargs):
+    def __init__(
+        self, decompiler_interface: DecompilerInterface = None, headless=False, auto_commit=True, reload_time=10,
+        do_safe_sync_all=False, **kwargs
+    ):
         self.headless = headless
         self.reload_time = reload_time
         if decompiler_interface is None:
@@ -141,8 +145,10 @@ class BSController:
             Comment: self.deci.comments,
             GlobalVariable: self.deci.global_vars,
             Enum: self.deci.enums,
+            Typedef: self.deci.typedefs,
             Struct: self.deci.structs,
-            Patch: self.deci.patches
+            Patch: self.deci.patches,
+            Segment: self.deci.segments
         }
 
         # client created on connection
@@ -166,7 +172,12 @@ class BSController:
 
         # create a pulling thread, but start on connection
         self._run_updater_threads = False
-        self.user_states_update_thread = threading.Thread(target=self.updater_routine)
+        self.user_states_update_thread = threading.Thread(target=self.updater_routine, daemon=True)
+
+        # other properties
+        self.progress_view_open = False
+        self.do_safe_sync_all = do_safe_sync_all
+        self.safe_synced_users = {}
 
         if self.headless:
             self._init_headless_components()
@@ -240,7 +251,9 @@ class BSController:
             self._ui_updater_worker.stop()
             self._ui_updater_thread.quit()
             _l.debug("Waiting for QThread ui_updater_thread to exit..")
-            self._ui_updater_thread.wait()
+            # TODO: on MacOS IDA Pro 8 >, this will hang the process, so we force the timeout after 5 seconds
+            #   this still causes a bad message for IDA Pro on Mac
+            self._ui_updater_thread.wait(2000)
 
     def schedule_job(self, cmd_func, *args, blocking=False, **kwargs):
         if not self._auto_commit_enabled:
@@ -284,6 +297,7 @@ class BSController:
 
             # do git pull/push operations if a remote exist for the client
             if self.client.last_pull_attempt_time is None:
+                _l.debug("Attempting to update states now")
                 self.client.commit_and_update_states(commit_msg="User created")
 
             # update every reload_time
@@ -312,6 +326,11 @@ class BSController:
                         Job(self._update_ui, all_states)
                     )
 
+                    # attempt to fast-sync anything that is easy to sync
+                    if self.do_safe_sync_all:
+                        self.safe_sync_all(all_states)
+
+
     def _update_ui(self, states):
         if not self.ui_callback:
             return
@@ -336,7 +355,6 @@ class BSController:
 
     def start_worker_routines(self):
         self._run_updater_threads = True
-        self.user_states_update_thread.daemon = True
         self.user_states_update_thread.start()
 
         self.push_job_scheduler.start_worker_thread()
@@ -445,7 +463,7 @@ class BSController:
         @param kwargs:
         @return:
         """
-        if self.is_not_syncing_data():
+        if self.should_watch_artifacts():
             self.commit_artifact(*args, **kwargs)
 
     @init_checker
@@ -484,6 +502,8 @@ class BSController:
         identifiers = DecompilerInterface.get_identifiers(artifact)
         current_art = get_art_func(state, *identifiers)
         merged_artifact = self.merge_artifacts(current_art, artifact, merge_level=MergeLevel.OVERWRITE)
+        if not merged_artifact:
+            return False
 
         # set the artifact in the target state, likely master
         _l.debug(f"Setting an artifact now into {state} as {artifact}")
@@ -521,6 +541,8 @@ class BSController:
             members=True,
             header=True,
             do_type_search=True,
+            do_merge=True,
+            fast_only=False,
             **kwargs
     ):
         state: State = state if state is not None else self.get_state(user=user, priority=SchedSpeed.FAST)
@@ -542,10 +564,21 @@ class BSController:
             # specify to BinSync that this is not user-changed, but merged from someone else
             target_artifact.reset_last_change()
 
-        merged_artifact = self.merge_artifacts(
-            master_artifact, target_artifact,
-            merge_level=merge_level, master_state=master_state
-        )
+        if do_merge:
+            merged_artifact = self.merge_artifacts(
+                master_artifact, target_artifact,
+                merge_level=merge_level, master_state=master_state
+            )
+        else:
+            merged_artifact = target_artifact
+
+        if merged_artifact is None:
+            self.deci.warning(
+                f"Failed to merge {master_artifact} with {target_artifact} "
+                f"using strategy {self.merge_level if merge_level is None else merge_level}."
+            )
+            return False
+
         if isinstance(merged_artifact, Struct):
             if merged_artifact.name.startswith("__"):
                 _l.info(f"Skipping fill for {target_artifact} because it is a system struct")
@@ -565,6 +598,13 @@ class BSController:
                 if do_type_search:
                     self.discover_and_sync_user_types(merged_artifact, state=state, master_state=master_state)
 
+                # in fast only, we disable writes that happen in the decompiler
+                if fast_only and isinstance(merged_artifact, Function):
+                    merged_artifact.stack_vars = {}
+                    if merged_artifact.header is not None:
+                        merged_artifact.header.args = {}
+                        merged_artifact.header.type = None
+
                 # set the imports into the decompiler
                 art_dict[identifier] = merged_artifact
 
@@ -578,7 +618,7 @@ class BSController:
                 fill_changes = False
                 _l.error(f"Failed to fill artifact {merged_artifact} because of an error {e}")
 
-        _l.info(
+        self.deci.info(
             f"Successfully synced new changes from {state.user} for {merged_artifact}" if fill_changes
             else f"No new changes or failed to sync from {state.user} for {merged_artifact}"
         )
@@ -615,16 +655,14 @@ class BSController:
         @return:
         """
         changes = False
-        # only do struct headers for circular references
         master_state, state = self.get_master_and_user_state(user=user, **kwargs)
+        art_dict = self.artifact_dict_map[Struct]
         for name, struct in state.structs.items():
+            # if we already synced a struct that was nested in a parent struct, skip it
+            if name in art_dict.keys() and art_dict[name] == struct:
+                continue
             changes |= self.fill_artifact(
-                name, artifact_type=Struct, user=user, state=state, master_state=master_state, members=False
-            )
-
-        for name, struct in state.structs.items():
-            changes |= self.fill_artifact(
-                name, artifact_type=Struct, user=user, state=state, master_state=master_state, header=False
+                name, artifact_type=Struct, user=user
             )
 
         return changes
@@ -658,6 +696,38 @@ class BSController:
 
         return changes
 
+    def fill_typedefs(self, user=None, do_type_search=True, **kwargs):
+        changes = False
+        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
+        for name, typedef in state.typedefs.items():
+            changes |= self.fill_artifact(
+                name, artifact_type=Typedef, user=user, state=state, master_state=master_state,
+                do_type_search=do_type_search
+            )
+
+        return changes
+
+    @init_checker
+    def fill_segment(self, name, user=None, do_type_search=True, **kwargs):
+        """Fill a single segment from another user."""
+        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
+        return self.fill_artifact(
+            name, artifact_type=Segment, user=user, state=state, master_state=master_state,
+            do_type_search=do_type_search
+        )
+
+    @init_checker
+    def fill_segments(self, user=None, do_type_search=True, **kwargs):
+        changes = False
+        master_state, state = self.get_master_and_user_state(user=user, **kwargs)
+        for name, segment in state.segments.items():
+            changes |= self.fill_artifact(
+                name, artifact_type=Segment, user=user, state=state, master_state=master_state,
+                do_type_search=do_type_search
+            )
+
+        return changes
+
     def sync_all(self, user=None, **kwargs):
         """
         Connected to the Sync All action:
@@ -671,7 +741,7 @@ class BSController:
 
         master_state, state = self.get_master_and_user_state(user=user, **kwargs)
         fillers = [
-            self.fill_enums, self.fill_global_vars, self.fill_functions
+            self.fill_enums, self.fill_global_vars, self.fill_functions, self.fill_segments
         ]
         changes = False
         # need to do structs specially
@@ -717,18 +787,21 @@ class BSController:
 
         # TODO: make structus work in IDA
         target_artifacts = target_artifacts or {
-            # Struct: self.fill_struct,
-            Comment: lambda *x, **y: None,
+            Struct: self.fill_artifact,
+            #Comment: lambda *x, **y: None,
+            Comment: self.fill_artifact,
             Function: self.fill_artifact,
             GlobalVariable: self.fill_artifact,
             Enum: self.fill_artifact
         }
+        total_synced = defaultdict(int)
 
         for artifact_type, filler_func in target_artifacts.items():
             _l.info(f"Magic Syncing artifacts of type {artifact_type.__name__} now...")
             pref_state = users_state_map[preference_user]
             for identifier in self.changed_artifacts_of_type(artifact_type, users=all_users + [preference_user],
                                                              states=users_state_map):
+                total_synced[artifact_type] += 1
                 pref_art = self.pull_artifact(artifact_type, identifier, state=pref_state)
                 for user in all_users:
                     user_state = users_state_map[user]
@@ -748,19 +821,156 @@ class BSController:
                     filler_func(
                         identifier, artifact_type=artifact_type, artifact=pref_art, state=master_state,
                         commit_msg=f"Magic Synced {pref_art}",
-                        merge_level=MergeLevel.NON_CONFLICTING
+                        merge_level=MergeLevel.NON_CONFLICTING,
+                        do_type_search=False
                     )
                 except Exception as e:
                     _l.info(f"Banishing exception: {e}")
 
         _l.info("Magic Syncing Completed!")
+        # summarize total synchage!
+        _l.info(f"In total: {total_synced[Struct]} Structs, {total_synced[Function]} Functions, "
+                f"{total_synced[GlobalVariable]} Global Variables, and {total_synced[Enum]} Enums were synced.")
+
+    def safe_sync_all(self, all_states: list[State]):
+        """
+        This function attempts to do the following:
+        - grab all the states for all users
+        - iterate all functions changed for all users, grab two things:
+            1. function name
+            2. function comments
+        - with all function names candidates, remove candidates for ones the master user already has a func name
+        """
+        addr_to_new_names_and_user = defaultdict(list)
+        master_state = next(state for state in all_states if state.user == self.client.master_user)
+        changes_per_func = self.compute_changes_per_function(states=all_states)
+        for state in all_states:
+            if state is master_state:
+                continue
+
+            for addr, light_func in state.functions.items():
+                name = light_func.name
+                if not name:
+                    continue
+
+                if self.is_default_name(name):
+                    continue
+
+                changes = changes_per_func.get(addr, {}).get(state.user, 0)
+                addr_to_new_names_and_user[addr].append((name, state.user, changes))
+
+        # now we have a list of functions that have new names, we want to rank the best one for each func addr
+        best_names_and_ids = {}
+        for addr, new_names in addr_to_new_names_and_user.items():
+            scored_candidates = [(name, user, BSController.readability_score(name), changes) for name, user, changes in new_names]
+            # max it by highest changes and then highest readability
+            best_candidate = max(scored_candidates, key=lambda x: (x[3], x[2]))
+            # name, user, changes num
+            best_names_and_ids[addr] = (best_candidate[0], best_candidate[1], best_candidate[2])
+
+        # now we have single name for each function, we for any we don't currently have
+        total_change = 0
+        for addr, (name, user, changes) in best_names_and_ids.items():
+            if user == self.client.master_user:
+                continue
+
+            fast_func = self.deci.fast_get_function(addr)
+            # the function must exist in the decompiler, and have a default name
+            if fast_func is None:
+                continue
+
+            update_cmt = ""
+            if addr in self.safe_synced_users:
+                old_name, old_user, old_change = self.safe_synced_users[addr]
+                if old_user == user and changes > old_change:
+                    update_cmt = f"[binsync]: {old_user} has {changes - old_change} new changes.\n"
+
+            needs_new_name = self.is_default_name(fast_func.name) and fast_func.name != name
+            if not needs_new_name and not update_cmt:
+                continue
+
+            # fill the function with the new name
+            succeed = False
+            if needs_new_name:
+                fast_func.name = name
+                succeed = self.fill_artifact(
+                    fast_func.addr, artifact_type=Function, artifact=fast_func, user=user, do_merge=False, fast_only=True
+                )
+            if update_cmt:
+                try:
+                    old_cmt: Comment = self.deci.comments[fast_func.addr]
+                except KeyError:
+                    old_cmt = Comment(addr=fast_func.addr, comment="")
+
+                if old_cmt and old_cmt.comment and "[binsync]" in old_cmt.comment:
+                    # remove the first line with the old comment
+                    old_cmt.comment = "\n".join(old_cmt.comment.split("\n")[1:])
+                    old_cmt.comment = update_cmt + old_cmt.comment
+                else:
+                    old_cmt.comment = update_cmt + old_cmt.comment
+
+                self.fill_artifact(
+                    fast_func.addr, artifact_type=Comment, artifact=old_cmt, user=user, do_merge=False, fast_only=True
+                )
+
+            self.safe_synced_users[addr] = (name, user, changes)
+            if succeed:
+                total_change += 1
+
+        if total_change:
+            _l.info(f"Safe Syncing completed with {total_change} changes.")
+
+    @staticmethod
+    def tokenize_varname(name):
+        """
+        Tokenizes a variable name by splitting on underscores and identifying camelCase boundaries.
+        Returns a list of lowercase tokens.
+        """
+        # Split on underscores
+        tokens = name.split('_')
+        camel_tokens = []
+        # Further split tokens with camelCase using a regex pattern.
+        for token in tokens:
+            parts = re.findall(r'[A-Z]?[a-z]+|[A-Z]+(?![a-z])', token)
+            if parts:
+                camel_tokens.extend(parts)
+            else:
+                camel_tokens.append(token)
+        return [tok.lower() for tok in camel_tokens if tok]
+
+    @staticmethod
+    def readability_score(var_name):
+        """
+        Calculates a readability score for the variable name based on word frequency.
+        Uses the logarithm of the word frequency probabilities (words not found get a small fallback value).
+        A higher score means the name looks more natural.
+        """
+        tokens = BSController.tokenize_varname(var_name)
+        score = 0.0
+        # For each token, look up the word frequency in English.
+        for token in tokens:
+            freq = word_frequency(token, 'en')
+            # Avoid log(0); if the token frequency is 0, assign a small fallback frequency.
+            if freq == 0:
+                freq = 1e-9
+            score += math.log(freq)
+        return score
+
+    @staticmethod
+    def is_default_name(name: str):
+        BAD_STARTS = ("sub_", "FUNC_")
+        for starter in BAD_STARTS:
+            if name.startswith(starter):
+                return True
+
+        return False
 
     #
     # Force Push
     #
 
     @init_checker
-    def force_push_functions(self, func_addrs: List[int]):
+    def force_push_functions(self, func_addrs: List[int], use_decompilation=False):
         """
         Collects the functions currently stored in the decompiler, not the BS State, and commits it to
         the master users BS Database. Function addrs should be in the lifted form.
@@ -768,16 +978,36 @@ class BSController:
         TODO: push the comments and custom types that are associated with each stack vars
         TODO: refactor to use internal push_function for correct commit message
         """
+
+        # NOTE: The following check allows to show a warning to a user and to
+        # avoid a ZeroDivisionError in the progress_bar declared later below
+        if not func_addrs:
+            _l.warning("Ignored force push, no function selected")
+            return
+
         master_state: State = self.client.master_state
         committed = 0
-        for func_addr in progress_bar(func_addrs, gui=not self.headless, desc="Decompiling functions to push..."):
-            f = self.deci.functions[func_addr]
-            if not f:
-                _l.warning(f"Failed to force push function @ {func_addr:#0x}")
-                continue
+        progress_str = "Decompiling functions to push..." if use_decompilation else "Collecting functions..."
 
-            master_state.functions[f.addr] = f
-            committed += 1
+        funcs = []
+        if use_decompilation:
+            for func_addr in progress_bar(func_addrs, gui=not self.headless, desc=progress_str):
+                f = self.deci.functions[func_addr]
+                if not f:
+                    _l.warning(f"Failed to force push function @ %s", func_addr)
+                    continue
+
+                funcs.append(f)
+        else:
+            # no progress bar needed!
+            _func_addrs = set(func_addrs)
+            for addr, func in self.deci.functions.items():
+                if addr in _func_addrs:
+                    funcs.append(func)
+
+        for func in funcs:
+            master_state.set_function(func)
+        committed += len(funcs)
 
         # commit the master state back!
         self.client.master_state = master_state
@@ -814,18 +1044,54 @@ class BSController:
         self.client.master_state = master_state
         self.deci.info(f"Globals force push successful: committed {committed} artifacts.")
 
+    @init_checker
+    def force_push_segments(self, segment_names: List[int]):
+        """
+        Collects the segments currently stored in the decompiler, not the BS State,
+        and commits them to the master users BS Database.
+
+        @param segment_names: List of segment names to push
+        @return: Success of committing the Segments
+        """
+        if not segment_names:
+            _l.warning("Ignored segments force push, no segments selected")
+            return
+
+        master_state: State = self.client.master_state
+        committed = 0
+        
+        for segment_name in segment_names:
+            try:
+                segment = self.deci.segments[segment_name]
+                if segment:
+                    master_state.set_segment(segment)
+                    committed += 1
+                else:
+                    _l.warning(f"Failed to force push segment @ {segment_name}")
+            except KeyError:
+                _l.warning(f"Segment at {segment_name} not found in decompiler")
+
+        self.client.master_state = master_state
+        self.deci.info(f"Segments force push successful: committed {committed} segments.")
+
     #
     # Utils
     #
 
-    def merge_artifacts(self, art1: Artifact, art2: Artifact, merge_level=None, **kwargs):
+    def merge_artifacts(self, art1: Artifact, art2: Artifact, merge_level=None, **kwargs) -> Optional[Artifact]:
         if merge_level is None:
             merge_level = self.merge_level
 
+        # error case
+        if art1 is None and art2 is None:
+            _l.warning("Attempting to merge two None artifacts, skipping...")
+            return None
+
+        # merge case does not matter if there is only the new artifact
         if art2 is None:
             return art1.copy()
-
-        if not art1 or (art1 == art2):
+        # always overwrite if the first artifact is None
+        if art1 is None or (art1 == art2):
             return art2.copy() if art2 else None
 
         if merge_level == MergeLevel.OVERWRITE or (not art1) or (art1 == art2):
@@ -905,47 +1171,62 @@ class BSController:
 
         return imported_types
 
-    def type_is_user_defined(self, type_str, state=None):
+    def type_is_user_defined(self, type_str, state=None) -> Tuple[Optional[str], Optional[Union[Struct, Enum, Typedef]]]:
         if not type_str:
-            return None
+            return None, None
 
         type_: CType = self.deci.type_parser.parse_type(type_str)
         if not type_:
             # it was not parseable
-            return None
+            return None, None
 
         # type is known and parseable
         if not type_.is_unknown:
-            return None
+            return None, None
 
         base_type_str = type_.base_type.type
-        return base_type_str if base_type_str in state.structs.keys() else None
+        # this could go wrong in overlapps of type names
+        for type_name, type_list in ((Struct, state.structs.keys()), (Enum, state.enums.keys()), (Typedef, state.typedefs.keys())):
+            if base_type_str in type_list:
+                return base_type_str, type_name
+        return None, None
 
     def sync_user_type(self, type_str, **kwargs):
         state = kwargs.pop('state')
         master_state = kwargs['master_state']
-        base_type_str = self.type_is_user_defined(type_str, state=state)
-        if not base_type_str:
+        base_type_str, base_type_cls = self.type_is_user_defined(type_str, state=state)
+        if base_type_str is None:
             return False
 
-        struct: Struct = state.get_struct(base_type_str)
-        if not struct:
-            return False
+        changes = False
+        if base_type_cls is Struct:
+            struct: Struct = state.get_struct(base_type_str)
+            if not struct:
+                return False
 
-        nested_undefined_structs = False
-        for off, memb in struct.members.items():
-            user_type = self.type_is_user_defined(memb.type, state=state)
-            if user_type and user_type not in master_state.structs.keys():
-                # should we ever happen to have a struct with a nested type that is
-                # also a struct that we don't have in our master_state, then we give up
-                # and attempt to fill all structs to resolve type issues
-                nested_undefined_structs = True
-                _l.info(f"Nested undefined structs detected, pulling all structs from {state.user}")
-                break
+            nested_undefined_structs = False
+            for off, memb in struct.members.items():
+                user_type, type_cls = self.type_is_user_defined(memb.type, state=state)
+                if type_cls is Struct and user_type not in master_state.structs.keys():
+                    # should we ever happen to have a struct with a nested type that is
+                    # also a struct that we don't have in our master_state, then we give up
+                    # and attempt to fill all structs to resolve type issues
+                    nested_undefined_structs = True
+                    _l.info(f"Nested undefined structs detected, pulling all structs from {state.user}")
+                    break
+
+            changes = self.fill_artifact(
+                base_type_str, artifact_type=Struct, state=state, **kwargs
+            ) if not nested_undefined_structs else self.fill_structs(state=state, **kwargs)
+        elif base_type_cls is Enum:
+            changes = self.fill_artifact(
+                base_type_str, artifact_type=Enum, state=state, **kwargs
+            )
+        elif base_type_cls is Typedef:
+            changes = self.fill_artifact(
+                base_type_str, artifact_type=Typedef, state=state, **kwargs
+            )
 
-        changes = self.fill_artifact(base_type_str, artifact_type=Struct, state=state,
-                                     **kwargs) if not nested_undefined_structs \
-            else self.fill_structs(state=state, **kwargs)
         return changes
 
     def get_master_and_user_state(self, user=None, **kwargs):
@@ -961,24 +1242,159 @@ class BSController:
     # Config Utils
     #
 
-    def load_saved_config(self):
+    def load_saved_config(self) -> Optional[BinSyncBSConfig]:
         config = BinSyncBSConfig().load_from_file()
         if not config:
-            return
-        self.config = config
-        _l.info(f"Loaded configuration file: '{self.config.save_location}'")
+            return None
 
         self.config = config
-        self.table_coloring_window = (config.table_coloring_window
-                                      or self.table_coloring_window)
+        _l.info(f"Loaded configuration file: '{self.config.save_location}'")
+        self.table_coloring_window = config.table_coloring_window or self.table_coloring_window
         self.merge_level = config.merge_level or self.merge_level
 
         if config.log_level == "debug":
             logging.getLogger("binsync").setLevel("DEBUG")
-            logging.getLogger("ida_binsync").setLevel("DEBUG")
-
         else:
             logging.getLogger("binsync").setLevel("INFO")
-            logging.getLogger("ida_binsync").setLevel("INFO")
 
         return self.config
+
+    #
+    # View Utils
+    #
+
+    def compute_changes_per_function(self, exclude_master=False, client=None, commit_hash=None, states=None):
+        """
+        Computes the number of changes per artifact type.
+        TODO: support more than just functions
+        TODO: make this not such a poormans counter. We should be using commits not this hack
+        """
+        if client is None:
+            client = self.client
+        before_ts = None
+        if commit_hash is not None:
+            try:
+                commit_ref = client.repo.commit(commit_hash)
+            except Exception as e:
+                _l.error(f"Failed to get commit {commit_hash} because of {e}")
+                commit_ref = None
+
+            if commit_ref is not None:
+                before_ts = commit_ref.committed_date
+
+        # gather all the states
+        if states:
+            all_states = states
+        else:
+            all_states = client.all_states(before_ts=before_ts)
+
+        if exclude_master:
+            all_states = [s for s in all_states if s.user != client.master_user]
+
+        func_addrs = list(self.deci.functions.keys())
+        function_counts = {addr: defaultdict(int) for addr in func_addrs}
+        for state in all_states:
+            for func_addr, func in state.functions.items():
+                func: Function
+                changes = 0
+                # check the parameters
+                header: FunctionHeader = func.header
+                if header:
+                    if header.name:
+                        changes += 1
+                    for arg in header.args.values():
+                        arg: FunctionArgument
+                        if arg.name:
+                            changes += 1
+                        if arg.type:
+                            changes += 1
+                    if header.type:
+                        changes += 1
+                # check the stack vars
+                for sv in func.stack_vars.values():
+                    sv: StackVariable
+                    if sv.name:
+                        changes += 1
+                    if sv.type:
+                        changes += 1
+
+                if func_addr not in function_counts:
+                    self.deci.warning(f"Function {func_addr} not found in the decompiler")
+                    function_counts[func_addr] = defaultdict(int)
+
+                function_counts[func_addr][state.user] = changes
+
+        return function_counts
+
+    def show_progress_window(self, *args, tag=None, **kwargs):
+        """
+        TODO: re-enable this later when you figure out how fix the apparent thread/proccess issue in IDA Pro
+        """
+
+        from binsync.ui.progress_graph.progress_window import ProgressGraphWidget
+
+        # TODO: XXX: re-enable this later
+        #if self.progress_view_open:
+        #    self.deci.info("Progress view already open")
+        #    return
+
+        self.deci.info("Collecting data to make progress view now...")
+        graph = self.deci.get_callgraph()
+        if tag == "master":
+            tag = None
+
+        self.deci.gui_attach_qt_window(ProgressGraphWidget, "Progress View", graph=graph, controller=self, tag=tag)
+        self.progress_view_open = True
+
+
+    def preview_function_changes(self, func_addr=None, user=None, **kwargs):
+        """
+        Get a preview of the function differences between two functions about to be synced.
+
+        This was written for the pop-up window that appears when hover over a sync or sync from in the 
+        function and context panel. Note that this only applies to functions and their comments. (Not for 
+        artifacts like structs etc). 
+
+        The approach to this is getting the two artifacts in question and creating a dictionary with information 
+        on name, type, args, and comments for master and target functions which can then be parsed to see 
+        how they differ. 
+        """
+        state = self.get_state(user=user, priority=SchedSpeed.FAST)
+        user = user or state.user
+        master_state = self.client.master_state
+        master_func = master_state.get_function(func_addr)
+        target_func = state.get_function(func_addr)
+        
+        # A lot of repetition so this is just a helper to get the relevant attributes 
+        def get_header_attr(func, attr):
+            return getattr(func.header, attr, None) if func and func.header else None
+        
+        # Get the comments where each comment is a dictionary 
+        get_comments = lambda state_obj: {addr: cmt.comment for addr, cmt in state_obj.get_func_comments(func_addr).items()}
+        # Need to handle the case that sometimes there is no master function in these cases 
+        master_comments = get_comments(master_state) if master_func else {}
+        target_comments = get_comments(state)
+        
+        diffs = {
+            'name': {
+                # can change this to use helper func since func and func.header should have same name 
+                'master': master_func.name if master_func else None,
+                'target': target_func.name if target_func else None
+            },
+            'args': {
+                'master': get_header_attr(master_func, 'args') or {},
+                'target': get_header_attr(target_func, 'args') or {}
+            },
+            'type': {
+                'master': get_header_attr(master_func, 'type'),
+                'target': get_header_attr(target_func, 'type')
+            },
+            'comments': {
+                'master': master_comments,
+                'target': target_comments
+            }
+        }
+
+        return diffs
+            
+
