@@ -21,7 +21,7 @@ from libbs.artifacts import (
     GlobalVariable,
     Patch,
     StackVariable,
-    Struct,
+    Struct, Typedef, Segment,
 )
 from libbs.artifacts import TomlHexEncoder
 from binsync import __version__ as BS_VERS
@@ -39,6 +39,7 @@ class ArtifactType:
     COMMENT = "comment"
     GLOBAL_VAR = "global variable"
     ENUM = "enum"
+    TYPEDEF = "typedef"
 
 
 #
@@ -52,6 +53,7 @@ def sanitize_name(unsafe_name: str) -> str:
     """
     return re.sub(r"[^a-zA-Z0-9_]", "_", unsafe_name)
 
+
 def update_dirty_flag(f):
     @wraps(f)
     def _update_dirty_flag(self, *args, **kwargs):
@@ -71,8 +73,8 @@ def update_last_change(f):
         artifact = args[0]
 
         # make a function if one does not exist
-        if isinstance(artifact, (FunctionHeader, StackVariable)):
-            func = self.get_or_make_function(artifact.addr)
+        func: Function | None = self.get_or_make_function(artifact.addr) \
+            if isinstance(artifact, (FunctionHeader, StackVariable)) else None
 
         if not should_set:
             from_user_msg = f" from {from_user}" if from_user else ""
@@ -81,67 +83,30 @@ def update_last_change(f):
 
         self.last_commit_msg = f"Updated {artifact}"
         artifact.last_change = datetime.datetime.now(tz=datetime.timezone.utc)
-
         # Comment
         if isinstance(artifact, Comment):
-            artifact_loc = artifact.addr
-            artifact_type = ArtifactType.COMMENT
-            # update function its in, if it's in a function
             func = self.find_func_for_addr(artifact.addr)
             if func:
                 func.last_change = artifact.last_change
 
         # Stack Var
         elif isinstance(artifact, StackVariable):
-            artifact_loc = artifact.addr
-            artifact_type = ArtifactType.FUNCTION
             func.last_change = artifact.last_change
 
-        elif isinstance(artifact, Function):
-            artifact_loc = artifact.addr
-            artifact_type = ArtifactType.FUNCTION
-
         # Function Header
         elif isinstance(artifact, FunctionHeader):
-            artifact_loc = artifact.addr
-            artifact_type = ArtifactType.FUNCTION
             func.last_change = artifact.last_change
 
-        # Patch
-        elif isinstance(artifact, Patch):
-            artifact_loc = artifact.offset
-            artifact_type = ArtifactType.PATCH
-
-        # Struct
-        elif isinstance(artifact, Struct):
-            artifact_loc = artifact.name
-            artifact_type = ArtifactType.STRUCT
-
-        # Global Var
-        elif isinstance(artifact, GlobalVariable):
-            artifact_loc = artifact.addr
-            artifact_type = ArtifactType.GLOBAL_VAR
-
-        # Enum
-        elif isinstance(artifact, Enum):
-            artifact_loc = artifact.name
-            artifact_type = ArtifactType.ENUM
-
-        else:
-            raise Exception("Undefined Artifact Type!")
-
-        self.last_push_artifact = artifact_loc
-        self.last_push_time = artifact.last_change
-        self.last_push_artifact_type = artifact_type
-
         return f(self, *args, **kwargs)
 
     return _update_last_change
 
 
 def list_files_in_dir(src: Union[pathlib.Path, git.Tree], dir_name, client=None) -> List[str]:
-    if client and isinstance(src, git.Tree):
-        files = client.list_files_in_tree(src)
+    from .client import Client
+
+    if isinstance(src, git.Tree):
+        files = Client.list_files_in_tree(src)
         return [name for name in files if name.startswith(dir_name)]
 
     # load from filesystem
@@ -159,8 +124,10 @@ def list_files_in_dir(src: Union[pathlib.Path, git.Tree], dir_name, client=None)
 
 
 def file_to_str(src: Union[pathlib.Path, git.Tree], filename, client=None) -> Optional[str]:
-    if client and isinstance(src, git.Tree):
-        file_data = client.load_file_from_tree(src, filename)
+    from .client import Client
+
+    if isinstance(src, git.Tree):
+        file_data = Client.load_file_from_tree(src, filename)
     else:
         if not src:
             src = pathlib.Path("")
@@ -219,6 +186,8 @@ class State:
         self.patches: Dict[int, Patch] = SortedDict()
         self.global_vars: Dict[int, GlobalVariable] = {}
         self.enums: Dict[str, Enum] = {}
+        self.typedefs: Dict[str, Typedef] = {}
+        self.segments: Dict[str, Segment] = {}
 
         # state is dirty on creation (metadata)
         self._dirty = dirty  # type: bool
@@ -230,12 +199,14 @@ class State:
                    and other.structs == self.structs \
                    and other.patches == self.patches \
                    and other.global_vars == self.global_vars \
-                   and other.enums == self.enums
+                   and other.enums == self.enums \
+                   and other.typedefs == self.typedefs \
+                   and other.segments == self.segments
         return False
 
     def copy(self):
         state = State(self.user, version=self.version, client=self.client, last_push_time=self.last_push_time, last_commit_msg=self.last_commit_msg, dirty=self._dirty)
-        artifacts = ["functions", "comments", "structs", "patches", "global_vars", "enums"]
+        artifacts = ["functions", "comments", "structs", "patches", "global_vars", "enums", "typedefs", "segments"]
         for artifact in artifacts:
             setattr(
                 state,
@@ -249,7 +220,8 @@ class State:
         return f"<State: {self.user} " \
                f"funcs={len(self.functions)} " \
                f"cmts={len(self.comments)} " \
-               f"globals={len(self.structs) + len(self.global_vars) + len(self.enums)}" \
+               f"globals={len(self.structs) + len(self.global_vars) + len(self.enums) + len(self.typedefs)} " \
+               f"segments={len(self.segments)}" \
                f">"
 
     def __repr__(self):
@@ -288,10 +260,10 @@ class State:
         d = {
             "user": self.user,
             "version": self.version,
-            "last_push_time": self.last_push_time,
-            "last_push_artifact": self.last_push_artifact,
-            "last_push_artifact_type": self.last_push_artifact_type,
-            "last_commit_msg": self.last_commit_msg,
+            #"last_push_time": self.last_push_time,
+            #"last_push_artifact": self.last_push_artifact,
+            #"last_push_artifact_type": self.last_push_artifact_type,
+            #"last_commit_msg": self.last_commit_msg,
         }
         self._dump_data(dst, 'metadata.toml', toml.dumps(d, encoder=TomlHexEncoder()).encode())
 
@@ -321,10 +293,12 @@ class State:
             self._dump_data(dst, path, struct.dumps(fmt=ArtifactFormat.TOML).encode())
 
         if pathlib.Path(self.client.repo_root + '/structs').exists():
+            sanitized_struct_names = set([sanitize_name(s_name) for s_name in self.structs.keys()])
             for path in pathlib.Path(self.client.repo_root + '/structs').iterdir():
                 file = path.stem
                 name = file.split(".")[0]
-                if name not in self.structs.keys():
+
+                if name not in sanitized_struct_names:
                     self._delete_data(dst, path)
 
         # dump comments
@@ -339,6 +313,12 @@ class State:
         # dump enums
         self._dump_data(dst, 'enums.toml', Enum.dumps_many(list(self.enums.values()), key_attr="name").encode())
 
+        # dump typedefs
+        self._dump_data(dst, 'typedefs.toml', Typedef.dumps_many(list(self.typedefs.values()), key_attr="name").encode())
+
+        # dump segments
+        self._dump_data(dst, 'segments.toml', Segment.dumps_many(list(self.segments.values()), key_attr="name").encode())
+
     @classmethod
     def parse(cls, src: Union[pathlib.Path, git.Tree], client=None):
         if isinstance(src, str):
@@ -377,6 +357,14 @@ class State:
         enums: List[Enum] = toml_file_to_artifacts(src, "enums.toml", Enum, client=client)
         state.enums = {enum.name: enum for enum in enums}
 
+        # load typedefs
+        typedefs: List[Typedef] = toml_file_to_artifacts(src, "typedefs.toml", Typedef, client=client)
+        state.typedefs = {typedef.name: typedef for typedef in typedefs}
+
+        # load segments
+        segments: List[Segment] = toml_file_to_artifacts(src, "segments.toml", Segment, client=client)
+        state.segments = {segment.name: segment for segment in segments}
+
         # load structs
         struct_files = list_files_in_dir(src, "structs", client=client)
         for struct_file in struct_files:
@@ -532,6 +520,32 @@ class State:
 
         return False
 
+    @update_dirty_flag
+    @update_last_change
+    def set_typedef(self, typedef: Typedef, set_last_change=True, **kwargs):
+        try:
+            old_typedef = self.typedefs[typedef.name]
+        except KeyError:
+            old_typedef = None
+
+        if old_typedef != typedef:
+            self.typedefs[typedef.name] = typedef
+            return True
+
+        return False
+
+    def set_segment(self, segment: Segment, set_last_change=True, **kwargs):
+        try:
+            old_segment = self.segments[segment.name]
+        except KeyError:
+            old_segment = None
+
+        if old_segment != segment:
+            self.segments[segment.name] = segment
+            return True
+
+        return False
+
     #
     # Getters
     #
@@ -646,6 +660,28 @@ class State:
     def get_enums(self):
         return self.enums
 
+    def get_typedef(self, name):
+        try:
+            typedef = self.typedefs[name]
+        except KeyError:
+            typedef = None
+
+        return typedef
+
+    def get_typedefs(self):
+        return self.typedefs
+
+    def get_segment(self, name):
+        try:
+            segment = self.segments[name]
+        except KeyError:
+            segment = None
+
+        return segment
+
+    def get_segments(self):
+        return self.segments
+
     def get_last_push_for_artifact_type(self, artifact_type):
         last_change = -1
         artifact = None
